
/* Common Open Systems Environment, Common Graphics Environment	 */		
/* Extension to	PEX 5.1	version	2.  File date: February	10, 1994 */		

/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/


 /* =======================================================================
  * NOTE NOTE NOTE  Insert this	section
  *		    at the end of pl_struct.c file
  * ====================================================================='=
  */

/***** Start of	Ext section ************/

#include "PEXExtlib.h"
#include "PEXExtlibint.h"


Status
PEXExtFetchElements (display, structure, whence1, offset1, whence2, offset2,
    float_format, numElementsReturn, sizeReturn, ocsReturn)

INPUT Display		*display;
INPUT PEXStructure	structure;
INPUT int		whence1;
INPUT long		offset1;
INPUT int		whence2;
INPUT long		offset2;
INPUT int		float_format;
OUTPUT unsigned	long	*numElementsReturn;
OUTPUT unsigned	long	*sizeReturn;
OUTPUT char		**ocsReturn;

{
    register pexExtFetchElementsReq	*req;
    char				*pBuf;
    pexExtFetchElementsReply		*rep;
    long				repSize;
    PEXExtOCData			*oc_data;
    int					server_float_format;
    int					escapeSize = 0;
    char				*escapeData;
    unsigned long			escapeOutDataSize;

    /*
     * Fill in the escape record.
     */

    escapeSize = SIZEOF	(pexExtFetchElementsReq);

    escapeData = PEXAllocBuf (escapeSize);

    server_float_format	= PEXGetProtocolFloatFormat (display);

    req	= (pexExtFetchElementsReq *) escapeData;

    req->fpFormat = server_float_format;
    req->sid = structure;
    req->position1_whence = whence1;
    req->position1_offset = offset1;
    req->position2_whence = whence2;
    req->position2_offset = offset2;

    /*
     * Generate	the escape.
     */

    rep	= (pexExtFetchElementsReply *) PEXEscapeWithReply (display,
	PEXExtEscapeOpcodeFetchElements, escapeSize, escapeData,
	&escapeOutDataSize);

    PEXFreeBuf (escapeData);

    if (rep == NULL) 
    {
	*sizeReturn = *numElementsReturn = 0;
	*ocsReturn = NULL;
	return (0);		/* return an error */
    }

    *numElementsReturn = rep->numElements;

    /* Start the pBuf after the	pad bytes in the reply data. */

    pBuf = (char *) rep	+ SIZEOF (pexExtFetchElementsReply);

    if (server_float_format != float_format)
    {
	/*
	 * Convert from	server's float format to the float format
	 * specified by	the application.
	 */

	oc_data	= PEXExtDecodeOCs (server_float_format,	rep->numElements,
	    (escapeOutDataSize - SIZEOF	(pexExtFetchElementsReply)), pBuf);

	*ocsReturn = PEXExtEncodeOCs (float_format, rep->numElements,
	    oc_data, sizeReturn);

	PEXExtFreeOCData (rep->numElements, oc_data);
    }
    else
    {
	/*
	 * No float conversion necessary.
	 */

	*sizeReturn = repSize =	escapeOutDataSize - 
	    SIZEOF (pexExtFetchElementsReply);
	if (*ocsReturn = (char *) PEXAllocBuf (repSize))
	    COPY_AREA (pBuf, *ocsReturn, (long)	repSize);
    }

    return (1);
}


Status
PEXExtFetchElementsAndSend (display, structure,
    whence1, offset1, whence2, offset2,	dstDisplay, resID, reqType)

INPUT Display		*display;
INPUT PEXStructure	structure;
INPUT int		whence1;
INPUT long		offset1;
INPUT int		whence2;
INPUT long		offset2;
INPUT Display		*dstDisplay;
INPUT XID		resID;
INPUT PEXOCRequestType	reqType;

{
    char				*pBuf;
    PEXDisplayInfo			*srcDisplayInfo;
    PEXDisplayInfo			*dstDisplayInfo;
    char				*ocAddr;
    PEXExtOCData			*oc_data;
    PEXEnumTypeDesc			*srcFloats;
    PEXEnumTypeDesc			*dstFloats;
    long				bytesLeft;
    int					getSize;
    int					size;
    unsigned long			oc_size;
    int					i, j;
    int					fp_match;
    int					float_format;
    unsigned long			numElementsReturn;
    unsigned long			sizeReturn;
    char				*ocsReturn;
    int					status;	      

    /*
     * Lock around the critical	section, for multithreading.
     */

    LockDisplay	(display);

    /*
     * Determine which floating	point format to	use.
     */

    PEXGetDisplayInfo (display,	srcDisplayInfo);
    PEXGetDisplayInfo (dstDisplay, dstDisplayInfo);

    if (srcDisplayInfo->fpFormat == dstDisplayInfo->fpFormat)
    {
	float_format = dstDisplayInfo->fpFormat;
	fp_match = 1;
    }
    else
    {
	srcFloats = srcDisplayInfo->fpSupport;
	dstFloats = dstDisplayInfo->fpSupport;

	fp_match = 0;
	for (i = 0; i <	dstDisplayInfo->fpCount	&& !fp_match; i++)
	    for	(j = 0;	j < srcDisplayInfo->fpCount; j++)
	    {
		if (dstFloats[i].index == srcFloats[j].index)
		{
		    float_format = dstFloats[i].index;
		    fp_match = 1;
		    break;
		}
	    }

	if (!fp_match)
	{
	    /*
	     * Will have to convert from source	display	float format to
	     * destination display float format.
	     */

	    float_format = srcDisplayInfo->fpFormat;
	}
    }

    UnlockDisplay (display);

    /*
     * Get the Elements.					  
     */

    status = PEXExtFetchElements (display, structure, whence1, offset1,	
	whence2, offset2, float_format,	&numElementsReturn, &sizeReturn, 
	&ocsReturn);		

    if (!status) 
    {
	PEXSyncHandle (display);
	return (0);		/* return an error */
    }

    /*
     * Lock around the critical	section, for multithreading.
     */

    LockDisplay	(display);

    /*
     * If no floating point conversion has to take place, fetch	the element
     * info directly into the destination display connection.
     */

    if (fp_match)
    {
	if (display == dstDisplay)
	    UnlockDisplay (display);

	if (PEXStartOCs	(dstDisplay, resID, reqType, float_format,
	    numElementsReturn, NUMWORDS(sizeReturn)))
	{
	    pBuf = (char *) ocsReturn;
	    bytesLeft =	sizeReturn;
	    getSize = PEXGetOCAddrMaxSize (dstDisplay);

	    while (bytesLeft > 0)
	    {
		size = bytesLeft < getSize ? bytesLeft : getSize;
		ocAddr = PEXGetOCAddr (dstDisplay, size);
		COPY_AREA (pBuf, ocAddr, (long)	size);
		pBuf +=	size;
		bytesLeft -= size;
	    }

	    PEXFinishOCs (dstDisplay);
	}

	if (display != dstDisplay)
	    UnlockDisplay (display);
    }
    else
    {
	/*
	 * Floating point conversion necessary.
	 */

	oc_data	= PEXExtDecodeOCs (float_format, numElementsReturn, sizeReturn,
	    ocsReturn);

	pBuf = PEXExtEncodeOCs (dstDisplayInfo->fpFormat, numElementsReturn,
	    oc_data, &oc_size);

	PEXExtFreeOCData (numElementsReturn, oc_data);

	if (display == dstDisplay)
	    UnlockDisplay (display);

	PEXSendOCs (dstDisplay,	resID, reqType,	dstDisplayInfo->fpFormat,
	    numElementsReturn, oc_size,	pBuf);

	if (display != dstDisplay)
	    UnlockDisplay (display);

    }

    PEXSyncHandle (dstDisplay);

    return (1);
}
/***** End of Ext section ************/
